/*! \file csmaMac.c
 \brief Carrier-Sensing Random Access MAC.
 
 @version 12
 @author Chris Hunter
 

 The csmaMac is a modified ALOHA MAC that
 serves as an example for novel MAC
 development. Nodes transmit whenever
 they have information to transmit, and only
 move on to the next packet once the original
 transmit is acknowledged (ACKed). If no ACK
 is received, a collision is inferred and the
 packet is re-transmitted.
 
 
 By default, the MAC also implements carrier-
 sensing multiple-access with collision-
 avoidance (CSMA-CA). This functionality is
 built into hardware peripherals in the project
 so very little software state is affected.
 
 In its current state, the project acts as
 a true ethernet MAC-level wireless bridge.
 Any ethernet activity that appears on one
 WARP will be sent to another via the custom
 wireless link.
 */

#include "warpmac.h"
#include "warpphy.h"
#include "csmaMac.h"
#include "xparameters.h"
#include "string.h"
#include "errno.h"
#include "stdlib.h"
#include "stdio.h"
#include "ascii_characters.h"
#include "warp_userioboard_util.h"

//Global variables to for user I/O board outputs
unsigned char userIOBoard_LEDs;
unsigned char charsToPrint[16];
unsigned int pktCount_good, pktCount_bad;

//Arrays to track pkt sequence numbers for each partner node
unsigned char rxSequences[16];
unsigned char txSequences[16];

unsigned char maximumReSend;

///MAC address table with agreed upon mapping between dipswitches and physical addresses
typedef struct {
	unsigned char addr[6];
} nodeAddr;

///Instance of the address table
nodeAddr nodeTable[16];

///Byte array containing physical MAC address of this node
unsigned char myAddr[6];
///Index to the routing table that identifies this node
unsigned char myID;

///Full rate modulation selection; QPSK by default
unsigned int pktFullRate;

//Payload code rate selection; must be CODE_RATE_NONE for uncoded PHY
unsigned int pktCodeRate;

///Buffer for holding a packet-to-xmit across multiple retransmissions
Macframe txBuffer;
///Buffer to hold received packet
Macframe rxBuffer;

///Current antenna selection
unsigned char currentAnt;
///Current 802.11 channel
unsigned char chan;

//Define handy macros for CSMA MAC packet types
///Data packet with payload meant for Ethernet transmission
#define DATAPACKET 1
///Acknowledgement packet meant for halting retransmissions
#define ACKPACKET 0

///@brief Callback for the depression of the left push button
///
///This function is empty by default
void left(){
}

///@brief Callback for the depression of the right push button
///
///This button switched between different fullrate modulation orders
void right(){
	switch(pktFullRate){
		case HDR_FULLRATE_BPSK:
			pktFullRate = HDR_FULLRATE_QPSK;
			xil_printf("QPSK\r\n");
			break;
		case HDR_FULLRATE_QPSK:
			pktFullRate = HDR_FULLRATE_QAM_16;
			xil_printf("16-QAM\r\n");
			break;
		case HDR_FULLRATE_QAM_16:
			pktFullRate = HDR_FULLRATE_QAM_64;
			xil_printf("64-QAM\r\n");
			break;
		case HDR_FULLRATE_QAM_64:
			pktFullRate = HDR_FULLRATE_BPSK;
			xil_printf("BPSK\r\n");
			break;
	}
}

///@brief Callback for the depression of the up push button
///
///This button increments the 2.4GHz channel being used; only valid channels (in [1,14]) will be used
void up(){
	unsigned int newFreq;
	
	chan = (chan > 13) ? 14 : chan+1;
	newFreq = warpphy_setChannel(GHZ_2, chan);
	warpmac_leftHex(chan);
	
	xil_printf("New Frequency %d\r\n", newFreq);
}

///@brief Callback for the depression of the middle push button
///
///This button decrements the 2.4GHz channel being used; only valid channels (in [1,14]) will be used
void middle(){
	unsigned int newFreq;
	
	chan = (chan < 2) ? 1 : chan-1;
	newFreq = warpphy_setChannel(GHZ_2, chan);
	warpmac_leftHex(chan);
	
	xil_printf("New Frequency %d\r\n", newFreq);
}


///@brief Updates the LCD
///
///Updates the LCD. This is only for the User I/O daughtercard. If you do not have this card,
///this function will do nothing. This operation takes some time, so it can only be called in a few places so as to
///not adversely affect the performance of the MAC.
void updateLCD()
{

	warp_userioboard_set_leds(userIOBoard_LEDs);
	
	snprintf(charsToPrint, 16, "Good: %09d ", pktCount_good);
	warp_userio_lcd_printline(charsToPrint, 16, 4, 1);
	
	snprintf(charsToPrint, 16, "Bad:  %09d ", (pktCount_bad)>>1);//bad pkts get counted twice
	warp_userio_lcd_printline(charsToPrint, 16, 5, 1);
	
	return;
}

///@brief Callback for the reception of UART bytes
///@param uartByte ASCII byte received from UART
///
///Provides the user with the bytes that was received over the serial port. This is useful for configuring
///PHY and MAC parameters in real time on a board.
void uartRecv_callback(unsigned char uartByte)
{
	if(uartByte != 0x0)
	{
		xil_printf("(%c)\t", uartByte);
		
		switch(uartByte)
		{
			case ASCII_1:
				pktFullRate = HDR_FULLRATE_BPSK;
				xil_printf("Tx Full Rate = BPSK\r\n");
				break;
				
			case ASCII_2:
				pktFullRate = HDR_FULLRATE_QPSK;
				xil_printf("Tx Full Rate = QPSK\r\n");
				break;
				
			case ASCII_4:
				pktFullRate = HDR_FULLRATE_QAM_16;
				xil_printf("Tx Full Rate = 16-QAM\r\n");
				break;
				
			case ASCII_6:
				pktFullRate = HDR_FULLRATE_QAM_64;
				xil_printf("Tx Full Rate = 64-QAM\r\n");
				break;
			case ASCII_D:
				warpphy_setPktDetMinDurPlus(1);
				break;
			case ASCII_d:
				warpphy_setPktDetMinDurMinus(1);
				break;
			case ASCII_P:
				warpphy_setPktDetPlus(100);
				break;
			case ASCII_p:
				warpphy_setPktDetMinus(100);
				break;
			case ASCII_F:
				if(chan<14) chan++;
				warpphy_setChannel(GHZ_2, chan);
				xil_printf("Current channel: %d\r\n",chan);
				break;
			case ASCII_f:
				if(chan>1) chan--;
				warpphy_setChannel(GHZ_2, chan);
				xil_printf("Current channel: %d\r\n",chan);
				break;
			case ASCII_C:
				warpphy_setCSMAPlus(100);
				break;
			case ASCII_c:
				warpphy_setCSMAMinus(100);
				break;
			case ASCII_A:
				warpphy_setSISOAntenna(1);
				xil_printf("Radio 3 Selected\r\n");
				break;
			case ASCII_a:
				warpphy_setSISOAntenna(0);
				xil_printf("Radio 2 Selected\r\n");
				break;
			case ASCII_G:
				warpphy_setTargetPlus();
				break;
			case ASCII_g:
				warpphy_setTargetMinus();
				break;
			default:
				xil_printf("Undefined command\r\n");
				break;
		}
	}
	
	return;
}
///@brief Callback for the expiration of timers
///
///This function is responsible for handling #TIMEOUT and #BACKOFF.
///The job responsibilities of this function are to:
///-increase the contention window upon the expiration of a #TIMEOUT
///-initiate a #BACKOFF timer upon the expiration of a #TIMEOUT
///-retransmit a packet upon the expiration of a #BACKOFF
///@param timerType #TIMEOUT or #BACKOFF
void timer_callback(unsigned char timerType){
	int status;
	switch(timerType){
		case TIMEOUT_TIMER:
			//warpmac_incrementLEDHigh();
			warpmac_setTimer(BACKOFF_TIMER);
			break;
		case BACKOFF_TIMER:
			//xil_printf("Remaining Retransmissions = %d\r\n",txBuffer.header.remainingReSend&0xF);
			if(txBuffer.header.remainingReSend&0xF){
				//Copy the header over to packet buffer 1
				warpmac_prepPhyForXmit(&txBuffer,1);
				//warpmac_setDebugGPIO(0xF);
				//Send packet buffer 1
				warpmac_startPhyXmit(1);
				//Wait for it to finish and enable the receiver
				warpmac_finishPhyXmit();
				//warpmac_setDebugGPIO(0);
				//Start a timeout timer
				//xil_printf("----SET TIMEOUT\r\n");
				warpmac_setTimer(TIMEOUT_TIMER);
				warpmac_decrementRemainingReSend(&txBuffer);
				//Copy the header over to packet buffer 1
				//warpmac_prepPhyForXmit(&txBuffer,1);
			}
			else warpmac_enableEthernet();
			break;
	}
}


///@brief Callback for the reception of Ethernet packets
///
///This function is called by the ethernet MAC drivers
///when a packet is available to send. This function fills
///the Macframe transmit buffer with the packet and sends
///it over the OFDM link
///@param length Length, in bytes, of received Ethernet frame
void emacRx_callback(Xuint32 length, char* payload){
	unsigned char rxSeqNum;
	unsigned char destNode;
	
	warpmac_resetCurrentCW();
	warpmac_disableEthernet();
	txBuffer.header.length = length;
	txBuffer.header.pktType = DATAPACKET;
	
	//Set the modulation scheme for the packet's full-rate symbols
	txBuffer.header.fullRate = pktFullRate;
	
	//Set the code rate for the packet's payload
	txBuffer.header.codeRate = pktCodeRate;
	
	//For now, assume our destination is our opposite ID (all traffic is 0 <-> 1)
	destNode = (myID+1)%2;
	
	//Use the next sequence number for this node (top four bits) and resend count of 0 (bottom four bits)
	txSequences[destNode] = txSequences[destNode]+0x10;
	txBuffer.header.remainingReSend = (((txSequences[destNode])&0xF0)|((maximumReSend+1)&0x0F));
	
	//Copy in the packet's destination MAC address
	//Hard-coded as this node's partner node
	memcpy(txBuffer.header.destAddr,nodeTable[destNode].addr,6);
	
	//If the medium is idle
	if(warpmac_carrierSense()){
		
		//Copy the header over to packet buffer 1
		warpmac_prepPhyForXmit(&txBuffer,1);
		//Send packet buffer 1
		
		
		warpmac_startPhyXmit(1);
		//Wait for it to finish and enable the receiver
		warpmac_finishPhyXmit();
		//warpmac_setDebugGPIO(0x0);
		//Start a timeout timer
		warpmac_setTimer(TIMEOUT_TIMER);
		warpmac_decrementRemainingReSend(&txBuffer);
		//Copy the header over to packet buffer 1
		//warpmac_prepPhyForXmit(&txBuffer,1);
	}
	else{
		//Start a backoff timer
		warpmac_setTimer(BACKOFF_TIMER);
	}
	
	return;
}

///@brief Callback for the reception of bad wireless headers
///
///@param packet Pointer to received Macframe
void phyRx_badHeader_callback() {
	warpmac_incrementLEDLow();
	
	userIOBoard_LEDs = userIOBoard_LEDs==0x80 ? 0x01 : (userIOBoard_LEDs << 1);

	pktCount_bad++;
	return;
}

///@brief Callback for the reception of good wireless headers
///
///This function then polls the PHY to determine if the entire packet passes checksum
///thereby triggering the transmission of the ACK and the transmission of the received
///data over Ethernet.
///@param packet Pointer to received Macframe
void phyRx_goodHeader_callback(Macframe* packet){
	
	Macframe ackPacket;
	unsigned char state=INCOMPLETE;
	unsigned char srcNode;
	unsigned char shouldSend = 0;
	
	
	//Exploit the fact that our MAC addresses use our ID as the last octet
	// This should be replaced by something more sophisticated at some point
	srcNode = packet->header.srcAddr[5];
	
	//Rotate the ring of LEDs clockwise
	userIOBoard_LEDs = userIOBoard_LEDs==0x1 ? 0x80 : (userIOBoard_LEDs >> 1);
	
	//If the packet is addressed to this node
	if(warpmac_addressedToMe(packet)){
		
		
		switch(packet->header.pktType){
				//If received packet is data
			case DATAPACKET:
				//Acknowledge the reception of the data
				
				ackPacket.header.length = 0;
				ackPacket.header.pktType = ACKPACKET;
				ackPacket.header.fullRate = pktFullRate;
				ackPacket.header.codeRate = pktCodeRate;
				
				memcpy(ackPacket.header.srcAddr,(unsigned char *)myAddr,6);
				memcpy(ackPacket.header.destAddr,packet->header.srcAddr,6);
				
				//Copy the header over to packet buffer 2
				warpmac_prepPhyForXmit(&ackPacket,2);
				
				//At this point, we have pre-loaded the PHY transmitter with the ACK in hoping that
				//the packet passes checksum. Now we wait for the state of the received to packet
				//to move from INCOMPLETE to either GOODPACKET or BADPACKET
				
				if((packet->header.length)>1400){
					updateLCD();
				}
				
				//state = warpmac_finishPhyRecv();
				state = warpmac_fastReply(2, (int)NULL, REPLYTOGOOD);
				
				warpmac_setDebugGPIO(0x0);
				if(state&GOOD){
					
					//Send packet buffer 2 containing the ACK
					
					//warpmac_startPhyXmit(2);
					//warpmac_incrementLEDHigh();
					
					//Update the right-hex display with the current resend count for the received pkt
					warpmac_leftHex(0xF & (packet->header.remainingReSend));
					
					//Check if the last received seq number for this partner node matches this received pkt
					// If not, record the new number and allow the pkt to be forwarded over the wire
					if(srcNode<16 && rxSequences[srcNode] != ((packet->header.remainingReSend)&0xF0)){
						rxSequences[srcNode] = (packet->header.remainingReSend)&0xF0;
						shouldSend = 1;	
					} 
					
					//Starts the DMA transfer of the payload into the EMAC
					if(shouldSend) warpmac_prepEmacForXmit(packet);
					
					//Blocks until the PHY is finished sending and enables the receiver
					warpmac_finishPhyXmit();
					
					//Waits until the DMA transfer is complete, then starts the EMAC
					if(shouldSend) warpmac_startEmacXmit(packet);
					
					pktCount_good++;
				}
				
				if(state&BAD){
					warpmac_incrementLEDLow();
					pktCount_bad++;
					
					//Rotate the ring of LEDs counter-clockwise
					userIOBoard_LEDs = userIOBoard_LEDs==0x80 ? 0x01 : (userIOBoard_LEDs << 1);

				}
				
				break;
				
			case ACKPACKET:
				//Clear the TIMEOUT and enable Ethernet
				if(warpmac_inTimeout()){
					warpmac_clearTimer(TIMEOUT_TIMER);
					warpmac_incrementLEDHigh();
					txBuffer.header.remainingReSend = 0;
					warpmac_resetCurrentCW();
					warpmac_setTimer(BACKOFF_TIMER);
				}
				
				pktCount_good++;
				
				break;
		}
	}
}

///@brief Main function
///
///This function configures MAC parameters, enables the underlying frameworks, and then loops forever.
int main(){		
	
	//Initialize global variables
	currentAnt = 0;
	chan = 9;
	
	pktCount_good = 0;
	pktCount_bad = 0;
	
	
	//Set the full-rate modulation to QPSK by default
	pktFullRate = HDR_FULLRATE_QPSK;
	
	//Set the payload coding rate to NONE by default
	//Note: Reference Design 12 PHY does not provide channel coding
	pktCodeRate = CODE_RATE_NONE;
	
	//Initialize the framework
	warpmac_init();
	maximumReSend = 8;
	warpmac_setMaxResend(maximumReSend);
	warpmac_setMaxCW(5);
	
	warpmac_setTimeout(100); //400
	warpmac_setSlotTime(22); //9 20
	
	//Read Dip Switch value from FPGA board.
	//This value will be used as an index into the routing table for other nodes
	myID = warpmac_getMyId();
	
	//Create an arbitrary address for this node
	// It is assumed throughout that the last octet matches the node's ID (its DIP switch)
	// This addressing scheme should be replaced with something more sophisticated someday
	unsigned char tmpAddr[6] = {0x16,0x24,0x63,0x53,0xe2,myID};
	
	//Copy the local MAC address to the global myAddr
	memcpy((unsigned char *)myAddr,(unsigned char *)tmpAddr,6);
	
	//Fill an arbitrary routing table so that nodes know each others' addresses
	unsigned char j;
	for(j=0;j<16;j++){
		nodeTable[j].addr[0] = myAddr[0];
		nodeTable[j].addr[1] = myAddr[1];
		nodeTable[j].addr[2] = myAddr[2];
		nodeTable[j].addr[3] = myAddr[3];
		nodeTable[j].addr[4] = myAddr[4];
		nodeTable[j].addr[5] = j;
	}
	
	//Sets the source MAC address
	warpmac_setMacAddr((unsigned char *)(&myAddr));
	
	//Rx buffer is where the EMAC will DMA Wireless payloads from
	warpmac_setRxBuffer(&rxBuffer,0);
	//Tx buffer is where the EMAC will DMA Ethernet payloads to
	warpmac_setTxBuffer(1);
	
	//Copy this node's MAC address into the Tx buffer's source address field
	memcpy(txBuffer.header.srcAddr,(unsigned char *)myAddr,6);
	
	//Register callbacks
	warpmac_setBadHeaderCallback((void *)phyRx_badHeader_callback);
	warpmac_setGoodHeaderCallback((void *)phyRx_goodHeader_callback);
	warpmac_setTimerCallback((void *)timer_callback);
	warpmac_setEmacCallback((void *)emacRx_callback);
	warpmac_setUpButtonCallback((void *)up);
	warpmac_setMiddleButtonCallback((void *)middle);
	warpmac_setLeftButtonCallback((void *)left);
	warpmac_setRightButtonCallback((void *)right);
	warpmac_setUartRecvCallback((void *)uartRecv_callback);
	
	userIOBoard_LEDs = 1;
	//Initialize the LCD for slot 1 (i.e. don't flip the image) and non-inverted colors
	warp_userioboard_lcd_init(1, 0);
	warp_userioboard_set_lcd_charbuf(1);
	
	//Print some static lines to the user I/O board LCD
	//snprintf(charsToPrint, 16, " Ref Design v12 ", pktCount_bad);
	warp_userio_lcd_printline("Ref Design v12.1", 16, 1, 1);
	//snprintf(charsToPrint, 16, "  Rx Pkt Counts", pktCount_bad);
	warp_userio_lcd_printline("  Rx Pkt Counts ", 16, 3, 1);
	
	//Set the user I/O board LEDs
	warp_userioboard_set_leds(userIOBoard_LEDs);
	
	
	//Use the radio in slot 2
	warpphy_setSISOAntenna(currentAnt);
	
	//Set the default center frequency
	warpphy_setChannel(GHZ_2, chan);
	
	//Enable carrier sensing
	warpmac_enableCSMA();
	
	//Enable Ethernet
	warpmac_enableEthernet();
	
	//Set the modulation scheme use for base rate (header) symbols
	warpmac_setBaseRate(QPSK);
	
	xil_printf("Reference Design v12.1\r\n");
	xil_printf("Beginning main loop\r\n");
	
	while(1){
		warpmac_pollPeripherals();
	}
	
	return;
}
