/*! \file warpphy.c
 \brief Provide PHY-specific functions for interfacing WARPMAC (and MAC code) to the OFDM PHY peripherals
 
 @version 12
 @author Chris Hunter & Patrick Murphy
 
 Many of the register names and bit assignments in this file depend on the design of the OFDM PHY. You must
 use matching versions of WARPMAC, WARPPHY and ofdm_txrx_mimo. Refer to the wireless reference designs for
 known-good combinations of versions.
 */

//XPS-generated header with peripheral parameters
#include "xparameters.h"

//Header for WARPPHY Interface
#include "warpphy.h"

//Header for WARPMAC framework (required for some global values)
#include "warpmac.h"

// These header files define some macros for reading/writing registers in the Sysgen-designed cores
// This layer of macros keep this code independent from the name of the sysgen core
//  (sysgen defines its register names as a function of the .mdl file name)
#include "ofdm_txrx_mimo_regMacros.h"
#include "ofdm_agc_mimo_regMacros.h"
#include "warp_timer_regMacros.h"

//Headers for other WARP peripheral cores
#include "EEPROM.h"
#include "radio_controller_basic.h"
#include "radio_controller_ext.h"
#include "radio_controller_adv.h"
#include "radio_controller_5ghz.h"

//Other standard header files
#include <stdio.h>
#include <string.h>


//********Globals********
int agc_targetRxPowerOut;
int agc_noisePowerEstimate;
unsigned int warpphy_sisoMode;
unsigned char baseRateMod;
unsigned int numTrainingSyms;
unsigned int ofdmRx_pktDetDly;
unsigned int numBaseRate;
unsigned char txGain;
unsigned char RxSoftDecodingEn;

unsigned int pktDet_energyThresh;
unsigned int pktDet_carrierSenseThresh;
unsigned char pktDet_threshMinDur;

unsigned int bothRadios = FIRST_RADIO | SECOND_RADIO;
unsigned int activeRadio, lastActiveRadio;

unsigned int A_KPval;
unsigned int B_KIval;
unsigned int B_KPval;
unsigned int PN_KVal;
unsigned int RxFFT_Window_Offset;
unsigned int txScaling_preamble;
unsigned int txScaling_payload;

/////////////////////////////////

///@brief Initializes the OFDM PHY core
///The OFDM PHY and radio controller cores have many options which must be configured before they can
///be used over the air. This funciton configures these cores with sensible defaults.
int warpphy_init(){

	xil_printf("  Initializing WARPPHY:\r\n");

	//Initialize global variables

	//AGC target output power, in units of dBv (voltage at MAX2829 output)
    agc_targetRxPowerOut = -15;

	//AGC noise power estimate, in dBm
	agc_noisePowerEstimate = -95;

	//SISO vs. MIMO mode
    warpphy_sisoMode = 1;

	//Base-rate modulation scheme; should be QPSK
    baseRateMod = BPSK;

	//Number of training symbol periods (all on 1 antenna for SISO, split across antennas for MIMO)
    numTrainingSyms = NUM_TRAINING_SYMBOLS;

	//Number of base rate OFDM symbols; every node must agree on this number ahead of time
    numBaseRate = NUM_BASERATE_SYMBOLS;

	//Packet detection delay, in sys_clk cycles
    // The Rx PHY waits this long before searching for correlation peaks
    ofdmRx_pktDetDly = 58;

    //Tx RF gain (6-bit value in [0,63])
	txGain = 0x3f;

    //Tx scaling values; both are interpreted as UFix16_13 values
    // The ratio of preamble/payload values should be approx 1/4
    // preamble=2480, payload=9792 tested well
    // preamble=2080, payload=8192 correspond to scaling by ~0.25/1.0
    txScaling_preamble = 2480;
    txScaling_payload  = 9792;
    
    //Packet detector parameters
	// Energy threshold - 6000/16 = RSSI of 375 (approx -75dBm Rx power)
    pktDet_energyThresh = 6000; //3400

	//Carrier sensing threshold - 8192/16 = RSSI of 512 (approx -65dBm Rx power)
    pktDet_carrierSenseThresh = 8192;

	//Minimum duration of energy-above-energyThresh required to assert a packet detection
    pktDet_threshMinDur = 48;
	
    //Set the default radio to the board in slot 2
	activeRadio = FIRST_RADIO;	

    //Define some carrier recovery filter parameters
	A_KPval = INIT_A_KPVAL;
	B_KIval = INIT_B_KIVAL;
	B_KPval = INIT_B_KPVAL;
	PN_KVal = INIT_PN_KVAL;

	//Define the FFT start offset (FFT will use (15 - offset) samles of cyclic prefix to handle synchronization
    RxFFT_Window_Offset = INIT_RXFFTOFSET;

	/*****************Radio Setup******************/
	xil_printf("	Initializing Radio Transmitter...\r\n");
	
	//The second argument sets the clock ratio in the radio controller's SPI controller
	// 1 is good for a 40MHz bus
	// 2 is the right value for an 80MHz bus
	WarpRadio_v1_Reset((unsigned int *)XPAR_RADIO_CONTROLLER_0_BASEADDR, 1);
	
	xil_printf("	Starting TxDCO correction...\r\n");

	//Apply the stored TX DC offset correction values for each radio
	warpphy_applyTxDCOCalibration(FIRST_RADIO);
	warpphy_applyTxDCOCalibration(SECOND_RADIO);
	
	//Configure some MAX2829 Tx parameters for good wideband performance

    //Set Tx bandwidth - 0x1 = 24MHz (minimum)
	WarpRadio_v1_TxLpfCornFreqCoarseAdj(0x1, FIRST_RADIO | SECOND_RADIO);
	
	//Enable hardware control of Tx gains (handled by the radio controller Tx state machine)
	WarpRadio_v1_SoftwareTxGainControl(0, FIRST_RADIO | SECOND_RADIO);
	
	//Four arguments: dly_TxEn, dly_TxStart, dly_GainRampEn, dly_PowerAmpEn, each in units of bus clock cycles
	WarpRadio_v1_SetTxTiming(FIRST_RADIO | SECOND_RADIO, 0, 200, 100, 0);

	//Three arguments: targetGain, gainStep, stepInterval; max targetGain is 0x3F
	WarpRadio_v1_SetTxGainTiming(FIRST_RADIO | SECOND_RADIO, txGain, 0xF, 1);
	
	//Set the MAX2829 Tx baseband gain; 0x3 is max gain
	WarpRadio_v1_BaseBandTxGain(0x3, FIRST_RADIO | SECOND_RADIO);
	
	xil_printf("	complete!\r\n");
	
	xil_printf("	Initializing Radio Receiver...");

	//Enable hardware AGC control of receive gains & RxHP
	WarpRadio_v1_RxHpSoftControlDisable(FIRST_RADIO | SECOND_RADIO);
	WarpRadio_v1_SoftwareRxGainControl(0, FIRST_RADIO | SECOND_RADIO);

	/*
	 //Bypass AGC - control Rx gain via software
	 WarpRadio_v1_RxHpSoftControlEnable(FIRST_RADIO | SECOND_RADIO);
	 WarpRadio_v1_SoftwareRxGainControl(1, FIRST_RADIO | SECOND_RADIO);
	 
	 WarpRadio_v1_RxLNAGainControl(0x2, FIRST_RADIO | SECOND_RADIO);
	 WarpRadio_v1_RxVGAGainControl(0x10, FIRST_RADIO | SECOND_RADIO);
	*/
	
	//Set bandwith with RxHP=0; RxHighPassCornerFreq(0) is critical for good performance
	WarpRadio_v1_RxHighPassCornerFreq(0, FIRST_RADIO | SECOND_RADIO);
	
	//Set Rx bandwidth; 0x0 = 15MHz (minimum)
	WarpRadio_v1_RxLpfCornFreqCoarseAdj(0, FIRST_RADIO | SECOND_RADIO);
	
	xil_printf("complete!\r\n");
	/**********************************************************/
	
	/*******************Packet Detector************************/
	xil_printf("	Initializing Packet Detection...");

	//Assert the packet detection output reset
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_ignoreDet(1);

	//Set the RSSI clock ratio; 0=sys_clk/4; 1=sys_clk/8
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setRSSIclkRatio(0);

    //Set the minimum energy duration
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setMinDuration(pktDet_threshMinDur);
	
    //Enable the carrier-sensing block's IDLE output (the timer can choose to use or ignore it)
	ofdm_txrx_mimo_WriteReg_Rx_CSMA_enableIdle(1);
	
    //Hold the packet detector's running sum in reset
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_resetSum(1);
	
	//Configure the packet detector's antenna masks (1=antA, 2=antB, 3=both)
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setMask(1);
	
	//Set the packet detector's logic mode for the two antenna ports; 0 = AND, 1 = OR
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setMode(1);
	
	//Set the running sum length; 16 is a good default
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setAvgLen(PKTDET_RSSI_SUMLEN);
	
	//Set the detector's energy threshold; this value must be scaled relative to the sum-length
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setThresh(pktDet_energyThresh);
	
	//Set the carrier sensing threshold to a very high level, disabling it by default
	ofdm_txrx_mimo_WriteReg_Rx_CSMA_setThresh(CSMA_DISABLED_THRESH);

	//Set the DIFS period; this is the require IDLE time before the IDLE output will assert
    // Units are sys_clk/4 cycles (10MHz), so 300 = 30us
    // This delay must exceed the TxDATA->TxACK turn around time (currently 23us)
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setDIFSPeriod(300);
	
	//Deassert the running sum reset; the sum will run continuously after this
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_resetSum(0);
	
	//Deassert the packet detection output reset; packet detection events will start after this
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_ignoreDet(0);
	
	xil_printf("complete!\r\n");
	/**********************************************************/
	
	/*************************PHY******************************/

	//Disable processing in the Tx PHY to prevent any accidental transmissions during setup
	mimo_ofdmTx_disable();

	//Disable processing in the Rx PHY to prevent any accidental receptions during setup
	mimo_ofdmRx_disable();

	xil_printf("	Initializing OFDM Transmitter...");
	
	//Configure the FFT scaling values in the PHY Tx and Rx
    // 6-bits each, 6LSB are Tx scaling; 6LSB+6 are Rx scaling
	mimo_ofdmTxRx_setFFTScaling((unsigned int)((
                                (16*RX_FFT_SCALING_STAGE1 + 4*RX_FFT_SCALING_STAGE2 + 1*RX_FFT_SCALING_STAGE3)<<6 ) | 
                                (16*TX_FFT_SCALING_STAGE1 + 4*TX_FFT_SCALING_STAGE2 + 1*TX_FFT_SCALING_STAGE3)));
	
	//Subcarrier indicies for pilot tones - the values here must line up with zeros in the modulation setup for each subcarrier
	// Subcarriers [-21,-7,7,21] are the default, matching the 802.11a spec
	mimo_ofdmTx_setPilot1Index((7 + ((64-7)<<16) ));
	mimo_ofdmTx_setPilot2Index((21 + ((64-21)<<16)));
	
	//Pilot tone values are Fix16_15, so
    // Pilot 1 should be negative, pilot 2 should be positive
	// 0x7FFF is +1, 0x8000 is -1
	// 0xA57D is ~-0.707, 0x5A82 is ~+0.707
	mimo_ofdmTx_setPilot1Value(0xA57D);
	mimo_ofdmTx_setPilot2Value(0x5A82); 
	
	//3 values in this write: number of training symbols, number of base rate symbols and number of full rate symbols
	// The number of full rate symbols is updated with each packet transmission, so it's set to a sensible non-zero value here
	warpphy_setNumSyms(numTrainingSyms + (numBaseRate*256) + (40*65536));
	
	//Scaling constant for the transmitted preamble; helps normalize the amplitude of the stored preamble and actual IFFT output
	mimo_ofdmTx_setTxScaling(txScaling_preamble, txScaling_payload);

	//Configure various options in the Tx PHY
	mimo_ofdmTx_setControlBits (
								(2 << 4) | //Preamble shift for antenna B
								(TX_SISO_MODE * warpphy_sisoMode) | //SISO mode
								TX_PILOT_SCRAMBLING | //Pseudo-random scrambling of pilot tones
								//TX_DISABLE_ANTB_PREAMBLE | //Disables preamble on antenna B
								//TX_SISO_ON_ANTB | //Uses antenna B for SISO mode
								0
								);
	
	xil_printf("complete!\r\n");
	
	//Finally enable the transmitter; it won't actually do anything until user code transmits a packet
	mimo_ofdmTx_enable();
	
	xil_printf("	Initializing OFDM Receiver...");

	//Configures the default number of samples of the cyclic prefix to use for synchronization offset tolerance
	// Larger values here reduce multipath tolerance
	mimo_ofdmRx_setFFTWindowOffset(INIT_RXFFTOFSET);
	
	//Scaling value is a 32-bit value, composed of two UFix16_11 values
	// So 0x08000800 scales A/B by exactly 1
	// This value is dependent on the number of training symbols
	// For SISO mode and numTraining=2, use 0x10001000 (the default)
	mimo_ofdmRx_setRxScaling(0x10001000); 
	
	//Long correlator parameters (for packet detection confirmation and symbol timing)
	// Top 16-bits are long-correlation threshold; 3000 works well over-the-air
	// Bottom 16-bits are sample index corresponding to the long correlation event
	//  (basically a magic number that aligns the Rx PHY state machine with the incoming packet)
	mimo_ofdmRx_setLongCorrParams(232  + (3000<<16));

	//Sets the delay between the pkt detection assertion and the PHY searching for correlation peaks
	mimo_ofdmRx_setPktDetDly(ofdmRx_pktDetDly);


	mimo_ofdmRx_setNumOFDMSyms(numTrainingSyms + (numBaseRate<<16));
	
	//Bottom 8 bits are the number of header bytes per packet; nodes must agree on this at build time
	//	bits[7:0]: number of header bytes (bytes in base-rate symbols)
	//Three more 8-bit values are stored here - each is an index of a byte in the header:
	//	bits[15:8]: less-significant byte of pkt length
	//	bits[23:16]: more-significant byte of pkt length
	//	bits[31:24]: dynamic modulation masks
	//	mimo_ofdmRx_setByteNums( (unsigned int)(NUM_HEADER_BYTES + (2<<8) + (3<<16) + (0<<24) ));
	mimo_ofdmRx_setByteNums( (unsigned int)(NUM_HEADER_BYTES + (3<<8) + (2<<16) + (0<<24) ));
	
	//Set filter coefficients for the CFO and phase noise tracking systems
	mimo_ofdmRx_setCFO_B_KI(INIT_B_KIVAL);
	mimo_ofdmRx_setCFO_B_KP(INIT_B_KPVAL);
	mimo_ofdmRx_setPNTrack_K(INIT_PN_KVAL);
	
	//Configure a bunch of options in the Rx PHY
	mimo_ofdmRx_setOptions
	(
		//RESET_BER |
		REQ_LONG_CORR |
		//BIG_PKTBUF_MODE |
		DYNAMC_PKT_LENGTHS |
		REQ_TWO_LONG_CORR |
		REQ_SHORT_CORR |
		(RX_SISO_MODE * warpphy_sisoMode) |
		CFO_USE_LONGCORR |
		DEBUG_CFO_OUTSEL |
		COARSE_CFO_EN | //Enable coarse CFO estimation
		USE_PILOT_ARCTAN |
		TX_DISABLE_PKTDET |
		//SWITCHING_DIV_EN | //Enable switching diversity at the receiver
		SIMPLE_DYN_MOD_EN |
		EXT_PKT_DETECT |
		//SISO_ON_ANTB | //Receive SISO stream on second radio
		//INT_PKT_DETECT |
		//EQ_BYPASS_DIVISION |
		RESET_ON_BAD_HDR | //Reset Rx PHY if header fails CRC
		0,
		DEFAULT_STATUSBITS
	 );
	
	warpphy_enableSisoMode();
	warpphy_setNumTrainingSyms(NUM_TRAINING_SYMBOLS);

	//Finally enable the Rx PHY
	mimo_ofdmRx_enable();
	
	//Set the modulation schemes (baseRate, antA_fullRate, antB_fullRate):
	// By default, we're in SISO mode, so antenna B gets 0 bits
	// 0xF enables dynamic modulation for antenna A (0xF gets AND'd with the header's full-rate field)
	warpphy_set_modulation(baseRateMod, 0xF, 0x0);
	
	//If you disable dynamic modulation, replace 0xF with an actual modulation scheme:
	//warpphy_set_modulation(QPSK, QPSK, 0x0);
	
	xil_printf("complete!\r\n");
	/**********************************************************/
	
	/***************************AGC****************************/
	xil_printf("	Initializing AGC...");
	ofdm_AGC_Initialize(agc_noisePowerEstimate);
	ofdm_AGC_setNoiseEstimate(agc_noisePowerEstimate);
	ofdm_AGC_SetDCO(1);
	ofdm_AGC_SetTarget(agc_targetRxPowerOut);
	ofdm_AGC_Reset();
	xil_printf("complete!\r\n");
	/**********************************************************/
	
	
	/**************************Timer***************************/
	//Initialize the timer hardware; the individual timers will get setup in WARPMAC
	warp_timer_init();
	/**********************************************************/
	
	//Finally, enable the Rx paths in the radios
	WarpRadio_v1_RxEnable(activeRadio);

#ifdef CONVCODED_PHY
	//Setup the PHY's encoder/decoder blocks
	//FIXME: the values below should be masked by their actual sizes
	ofdm_txrx_mimo_WriteReg_FEC_Coding(OFDM_BASEADDR,
		(SOFT_DECODING * RxSoftDecodingEn) |
		(QPSK_SCL * 2^4) |
		(QAM16_SCL * 2^8)
		);
#endif	

	return 0;
}

///@brief Clears any pending Rx pkt status in the PHY. Was warpphy_pktAck() in previous versions.
///
///The Rx PHY blocks after asserting either its good or bad packet status bits. The bits
///are cleared by asserting then de-asserting a register bit per status bit. This funciton clears
///both good and bad pkt bits; there is no harm in clearing a status bit that isn't actually asserted.
void warpphy_clearRxPktStatus(){
	
	//The TxRx_Interrupt_PktBuf_Ctrl register has many bits.
	// bits[2:0] are the Rx status enables
	// bits[6:4] are the Rx status resets
	// bit[3] and bit[7] are used for the TxDone status and are preserved here
	ofdm_txrx_mimo_WriteReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR) | DEFAULT_STATUSBITRESETS);
	ofdm_txrx_mimo_WriteReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR) &  ~DEFAULT_STATUSBITRESETS);

	return;
}

///@brief Releases the OFDM Rx PHY master reset
///
///De-asserting RX_GLOBAL_RESET allows the Rx PHY to begin processing packets.
void mimo_ofdmRx_enable(){
	
	//Clear any stale interrupts; this should never really be required
	warpphy_clearRxPktStatus();
	
	//De-asert the global reset
	ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR) & ~RX_GLOBAL_RESET);
	
	return;
}

///@brief Holds the OFDM Rx PHY in reset
///
///Asserting the RX_GLOBAL_RESET bit clears nearly all the state in the OFDM Rx. Configuration registers are not cleared.
void mimo_ofdmRx_disable(){
	
	//Assert the global reset
	ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR) | RX_GLOBAL_RESET);

	//Clear any stale interrupts; this should never really be required
	warpphy_clearRxPktStatus();
	
	return;
}

///@brief Configures options in the Rx PHY
///
///@param someOptions 32-bit options value, composed on bitwise OR'd values from warpphy.h
///@param intType Selects whether to interrupt on good or bad packets. Bitwise OR'd combination of INTR_BAD_PKTS and  INTR_GOOD_PKTS
void mimo_ofdmRx_setOptions(unsigned int someOptions, unsigned int blockingStatusBits){
	
	//Write the full controlBits register
	ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, someOptions);
	
	//The interrupt control bits are in the Interrupt_PktBuf_Ctrl register - bits[7:4]
	//Clear the interrupt control bits
	ofdm_txrx_mimo_WriteReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR) & 0xFFFFFF00);
	
	//Write just the interrupt control bits
	ofdm_txrx_mimo_WriteReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR) | (ALL_STATUSBITS_ENABLE & blockingStatusBits));
	
	return;
}

///@brief Returns the current value of the Rx PHY configuration register
///
/// Returns the value of the OFDM Rx ControlBits register. Use the bit masks from warpphy.h to decode the indivitual bits.
unsigned int mimo_ofdmRx_getOptions(){
	
	return ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR);
}


///@brief Holds the OFDM Tx in reset
///
/// Holds the OFDM Tx in reset; this prevents any state changes in the Tx PHY. Configuration registers are not affected.
void mimo_ofdmTx_disable(){
	
	//Assert the OFDM Tx master reset and pktDone reset
	ofdm_txrx_mimo_WriteReg_Tx_Start_Reset_Control(OFDM_BASEADDR, TX_MASTER_RESET);
	
	return;
}

///@brief Releases the OFDM Tx reset
///
/// Releases the OFDM Tx reset
void mimo_ofdmTx_enable(){
	
	//Assert then clear the OFDM Tx master reset and pktDone reset
	ofdm_txrx_mimo_WriteReg_Tx_Start_Reset_Control(OFDM_BASEADDR, TX_MASTER_RESET);
	ofdm_txrx_mimo_WriteReg_Tx_Start_Reset_Control(OFDM_BASEADDR, 0x0);
	
	return;
}


///@brief Initiates the transmission of a packet
///
///Starts the transmission of a packet from the OFDM Tx PHY. If blocking is enabled, this function returns only
///after the transmission finishes. In this mode, the radio receiver is automatically re-enabled. If blocking is disabled,
///the receiver must be re-enabled in user code later.
///
///@param block Selects whether this funciton blocks until the transmission finishes; use TXBLOCK or TXNOBLOCK
///@param radio Selects which radios are enabled for the transmission
inline int warpphy_pktTx(unsigned int block){
	
	
	//Hold the packet detector running RSSI sum in reset; this guarantees no packet detection events
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_resetSum(1);
	
	//First checks whether the Tx PHY is already transmitting a packet
	// If so, give up and return
/*	if(ofdm_txrx_mimo_ReadReg_Tx_PktRunning(OFDM_BASEADDR) == 1)
	{
		//Tx PHY was already busy sending a packet, which should be impossible
		xil_printf("Tx PHY was already transmitting! Failing...\r\n");
		
		// Return with a failure
		return 1;
	}*/
	
	/* Should there be a check here whether the Rx PHY is receiving a packet?
	 And what should happen if it is? It's sort of carrier sensing, but if
	 CSMA is disabled, should the Tx pkt be discarded, or held until the Rx
	 is no longer busy...
	 */
	
//	warpmac_setDebugGPIO(0xF);
	
	WarpRadio_v1_TxEnable(activeRadio);
	
//	warpmac_setDebugGPIO(0x0);

	//Sleep long enough for the PHY to start transmitting; depends on second argument to SetTxTiming
	usleep(6);
	
	//Return successfully
	return 0;
}

///@brief Polls PHY transmitter and re-enables reception upon completion
///
///This function blocks until the transmitter is complete and then re-enables
///reception by enabing the radio receiver and enabling packet detection.
int warpphy_waitForTx(){

	//warpmac_setDebugGPIO(0xF);

	//Poll the PHY transmitter until it's finished transmitting
	while(ofdm_txrx_mimo_ReadReg_Tx_PktRunning(OFDM_BASEADDR) == 1) {
		//warpmac_incrementLEDLow();
	}

	//warpmac_setDebugGPIO(0x0);

	//Transmission is done; re-enable the radio receiver
	WarpRadio_v1_RxEnable(activeRadio);

	//De-assert the packet detector reset; the running sum starts again and will begin detecting packets
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_resetSum(0);

	return 0;
}


///@brief Sets the packet buffer indicies for the OFDM Tx and Rx PHY.
///
///The PLB_BRAM used a the PHY packet buffer is large enough to hold many PHY packets.
///This BRAM is divided into many sub-buffers; the PHY can be set to use any sub-buffer for Tx or Rx.
///
///@param txBufOffset 6-bit integer selecting the sub-buffer for the PHY Tx
///@param rxBufOffset 6-bit integer selecting the sub-buffer for the PHY Rx
void warpphy_setBuffs(unsigned char txBufOffset, unsigned char rxBufOffset){
	
	//TxRx_Interrupt_PktBuf_Ctrl[21:16] compose the Rx buffer offset
	//TxRx_Interrupt_PktBuf_Ctrl[13:8]  compose the Tx buffer offset

	//First, zero out the current pkt buff offsets
	//Preserve the bottom 8 bits of the register (used for interrupt control)
	ofdm_txrx_mimo_WriteReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR) & 0x000000FF);
	
	//Then write the new pkt buff offsets
	ofdm_txrx_mimo_WriteReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR,
		ofdm_txrx_mimo_ReadReg_TxRx_Interrupt_PktBuf_Ctrl(OFDM_BASEADDR) |
		( (txBufOffset & 0x3F) << 8 ) |
		( (rxBufOffset & 0x3F) << 16 )
	);

	return;
}

///@brief Configures the PHY for SISO mode
void warpphy_enableSisoMode(){
	
	//Disable the TxRx paths during the switch
	WarpRadio_v1_TxRxDisable(FIRST_RADIO | SECOND_RADIO);

	//Update the global variable
	warpphy_sisoMode = 1;
	
	//Disable full-rate modulation for second anteanna
	warpphy_set_modulation(baseRateMod, 0xF, 0);

	//Update the Rx scaling
	// Assumes numTraining = 2
	mimo_ofdmRx_setRxScaling(0x10001000); 

	//Enable only one radio
	activeRadio = FIRST_RADIO;
	
	//Set SISO mode in the Tx and Rx PHY registers
	mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() | TX_SISO_MODE);
	mimo_ofdmRx_setOptions(mimo_ofdmRx_getOptions() | RX_SISO_MODE, DEFAULT_STATUSBITS);

	//Re-enable reception on just the first radio
	WarpRadio_v1_RxEnable(activeRadio);
	
	return;
}

///@brief Configures the PHY for MIMO mode
void warpphy_enableMimoMode(){

	//Disable the TxRx paths during the switch
	WarpRadio_v1_TxRxDisable(FIRST_RADIO | SECOND_RADIO);

	//Update the global variable
	warpphy_sisoMode = 0;
	
	//Enable full-rate modulation for both anteannas
	warpphy_set_modulation(baseRateMod, 0xF, 0xF);

	//Update the Rx scaling
	// Assumes numTraining = 2
	mimo_ofdmRx_setRxScaling(0x08000800); 

	//Enable both radios
	activeRadio = (FIRST_RADIO | SECOND_RADIO);
	
	//Un-set SISO mode in the Tx and Rx PHY registers
	mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() & (~TX_SISO_MODE));
	mimo_ofdmRx_setOptions(mimo_ofdmRx_getOptions() & (~RX_SISO_MODE),  DEFAULT_STATUSBITS);

	//Re-enable reception on both radios
	WarpRadio_v1_RxEnable(activeRadio);

	return;
}

///@brief Configures the TX PHY for MISO mode
void warpphy_enableMisoMode(){
	WarpRadio_v1_TxRxDisable(FIRST_RADIO | SECOND_RADIO);
	warpphy_sisoMode = 0;
	lastActiveRadio = activeRadio;
	activeRadio = FIRST_RADIO | SECOND_RADIO;
	mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() & (~TX_SISO_MODE));
	WarpRadio_v1_RxEnable(activeRadio);
}

///@brief Configures the TX PHY for SISO mode
void warpphy_disableMisoMode(){
	WarpRadio_v1_TxRxDisable(FIRST_RADIO | SECOND_RADIO);
	warpphy_sisoMode = 1;
	activeRadio = lastActiveRadio;
	mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() | TX_SISO_MODE);
	WarpRadio_v1_RxEnable(activeRadio);
}

///@brief Sets the number of training symbol periods used per packet
///
///Configures the number of training symbols which are transmitted with each packet. The Tx and Rx nodes must be
///configured for the same number. In SISO mode, the single channel is trained c times. In MIMO mode, each channel
///is trained c/2 times
///
///@param c number of training periods; must be even
void warpphy_setNumTrainingSyms(unsigned int numTraining){
	//Update the global variable; used each time a packet is transmitted
	numTrainingSyms = numTraining;
	
	//Configure the receiver
	mimo_ofdmRx_setNumOFDMSyms(numTrainingSyms + (numBaseRate<<16));
}

///@brief Configure the flexible modulation/demodulation in the OFDM PHY
///
///The OFDM PHY supports flexible modulation, allowing any combination of schemes per subcarrier
///Currently this code supports simple dynamic modulation, with 48 of 64 subcarriers assigned to carry data, 4 for pilots and 12 empty.
///The modulation scheme in the 48 data subcarriers is set by this funciton.
///
///@param baseRate Modulation scheme for base rate symbols
///@param antAFullRate Modulation scheme for full rate symbols on antenna A
///@param antBFullRate Modulation scheme for full rate symbols on antenna B
void warpphy_set_modulation(unsigned char baseRate, unsigned char antAFullRate, unsigned char antBFullRate)
{
#ifdef CONVCODED_PHY
	xil_printf("warpphy_set_modulation: Flexible modulation not currently supported in coded PHY! Giving up...\r\n");
	return;
#else
	unsigned int modIndex;

	//Update the global baseRate modulation variable
	baseRateMod = baseRate;
	
	//Define the standard subcarrier mapping - 48 used for data (0xF here), 4 pilots & 12 unused (0x0 here)
	// The vector contains 192 elements:
	//    0:63 - Antenna A full rate masks for subcarriers [0,1,2,...31,-31,-30,...-1]
	//  64:127 - Antenna B full rate masks for subcarriers [0,1,2,...31,-31,-30,...-1]
	// 128:191 - Base rate masks for subcarriers [0,1,2,...31,-31,-30,...-1]
	//The default masks are 3 copies of the same 64-length vector; yes, it's inefficient, but this scheme maintains flexibility in changing the mapping per antenna
	unsigned char modMasks[192] = {
			0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF,
			0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF,
			0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF, 0x0, 0xF, 0xF, 0xF, 0xF, 0xF, 0xF
	};

	//The PHY's shared memories for modulation masks have 192 4-bit entries; each entry's address is 4-byte aligned (hence the *sizeof(int) below )
	//Configure Tx and Rx antenna A full rate
	 for(modIndex=0; modIndex<64; modIndex++)
	 {
		XIo_Out32(XPAR_OFDM_TXRX_MIMO_PLBW_0_MEMMAP_TXMODULATION+(modIndex*sizeof(int)), modMasks[modIndex] & antAFullRate);
		XIo_Out32(XPAR_OFDM_TXRX_MIMO_PLBW_0_MEMMAP_RXMODULATION+(modIndex*sizeof(int)), modMasks[modIndex] & antAFullRate);
	 }

	 //Configure Tx and Rx antenna B full rate
	 for(modIndex=64; modIndex<128; modIndex++)
	 {
		XIo_Out32(XPAR_OFDM_TXRX_MIMO_PLBW_0_MEMMAP_TXMODULATION+(modIndex*sizeof(int)), modMasks[modIndex] & antBFullRate);
		XIo_Out32(XPAR_OFDM_TXRX_MIMO_PLBW_0_MEMMAP_RXMODULATION+(modIndex*sizeof(int)), modMasks[modIndex] & antBFullRate);
	 }
	 
	 //Configure the Tx and Rx base rate
	 for(modIndex=128; modIndex<192; modIndex++)
	 {
		XIo_Out32(XPAR_OFDM_TXRX_MIMO_PLBW_0_MEMMAP_TXMODULATION+(modIndex*sizeof(int)), modMasks[modIndex] & baseRate);
		XIo_Out32(XPAR_OFDM_TXRX_MIMO_PLBW_0_MEMMAP_RXMODULATION+(modIndex*sizeof(int)), modMasks[modIndex] & baseRate);
	 }

	return;
#endif
}

///@brief Selects which antenna is used when the PHY is in SISO mode
///
/// The OFDM PHY can use either antenna when in SISO mode. By default, antenna A is used.
///
///@param anteSel 0: use Antenna A, 1: use Antenna B
void warpphy_setSISOAntenna(unsigned char antSel)
{
	
	if(warpphy_sisoMode != 1) return; //Not valid if using MIMO mode
	if(mimo_ofdmRx_getOptions()  & SWITCHING_DIV_EN) return; //Not valid if using switching diversity
	
	//Disbale both transceivers during the switch
	WarpRadio_v1_TxRxDisable(bothRadios);

	//Temporarily disable packet detections
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_ignoreDet(1);

	if(antSel == 0) //choose radio in slot 2
	{
		//Select the raido in slot 2
		activeRadio = FIRST_RADIO;
		
		//Switch the pkt detector to listen to just radio #2
		ofdm_txrx_mimo_WriteReg_Rx_PktDet_setMask(1);
		
		//Disable antB selection in the PHY
		mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() & (~TX_SISO_ON_ANTB) );
		mimo_ofdmRx_setOptions(mimo_ofdmRx_getOptions() & (~SISO_ON_ANTB), DEFAULT_STATUSBITS);
	}
	else if(antSel == 1) //chose radio in slot 3
	{
		//Select the raido in slot 3		
		activeRadio = SECOND_RADIO;
		
		//Switch the pkt detector to listen to just radio #3
		//ofdm_pktDetector_mimo_WriteReg_pktDet_detectionMask(PKTDET_BASEADDR, 2);
		ofdm_txrx_mimo_WriteReg_Rx_PktDet_setMask(2);
		
		//Disable antB selection in the PHY
		mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() | (TX_SISO_ON_ANTB) );
		mimo_ofdmRx_setOptions(mimo_ofdmRx_getOptions() | (SISO_ON_ANTB), DEFAULT_STATUSBITS);
	}

	//Re-enable the receiver
	WarpRadio_v1_RxEnable(activeRadio);

	//Turn packet detection back on
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_ignoreDet(0);
	
	return;
}

///@brief Set the center frequency of the radio transceivers
///@param band Selects 2.4GHz or 5GHz bands (using GHZ_2 or GHZ_5)
///@param c Selects the channel number in the chosen band
///@return Returns -1 if an invalid band or channel was specified; otherwise returns the new center frequency in MHz
int warpphy_setChannel(unsigned char band, unsigned int chan){
	int newFreq = -1;
	
	if (band == GHZ_2){
		newFreq = WarpRadio_v1_SetCenterFreq2GHz(chan, FIRST_RADIO | SECOND_RADIO);
	}
	if (band == GHZ_5){
		newFreq = WarpRadio_v1_SetCenterFreq5GHz(chan, FIRST_RADIO | SECOND_RADIO);
	}
		
	return newFreq;
}

///@brief Set the center frequency of the radio transceivers independently
///@param antA_band Selects 2.4GHz or 5GHz bands (using GHZ_2 or GHZ_5) for antenna A
///@param antB_band Selects 2.4GHz or 5GHz bands (using GHZ_2 or GHZ_5) for antenna B
///@param antA_chan Selects the channel number in the chosen band for antenna A
///@param antB_chan Selects the channel number in the chosen band for antenna B
///@return Returns -1 if an invalid band or channel was specified; otherwise returns the new center frequency for antenna A in MHz
int warpphy_setSeparateChannels(unsigned char antA_band, unsigned int antA_chan, unsigned char antB_band, unsigned int antB_chan){

	int newFreq_A = -1;
	int newFreq_B = -1;
	
	if (antA_band == GHZ_2)
		newFreq_A = WarpRadio_v1_SetCenterFreq2GHz(antA_chan, FIRST_RADIO);
	if (antB_band == GHZ_2)
		newFreq_B = WarpRadio_v1_SetCenterFreq2GHz(antB_chan, SECOND_RADIO);
	if (antA_band == GHZ_5)
		newFreq_A = WarpRadio_v1_SetCenterFreq5GHz(antA_chan, FIRST_RADIO);
	if (antB_band == GHZ_5)
		newFreq_B = WarpRadio_v1_SetCenterFreq5GHz(antB_chan, SECOND_RADIO);
	if(newFreq_A == -1 || newFreq_B == -1)
		return -1;
	else
		return newFreq_A;
}



///@brief Applies TX DC offset calibration to the specified radios; the calibration values are stored in the radio's EEPROM
///@param radioSelection OR'd combinaton of RADIOx_ADDR values, specifying which radios to update
///@return Returns -1 if an EEPROM error occurs; returns 0 if successful
int warpphy_applyTxDCOCalibration(unsigned int radioSelection)
{
	int eepromStatus = 0;
	short calReadback = 0;
	signed short best_I, best_Q;
	unsigned char radioNum;
	Xuint8 memory[8], version, revision, valid, MAC[6], i;
	Xuint16 serial;

	//Radio selection will be 0x11111111, 0x22222222, 0x44444444 or 0x88888888
	// corresponding to radios in slots 1, 2, 3 or 4
	// We need the slot number to initialize the EEPROM
	radioNum = (radioSelection & 0xF) == 1 ? 1 : ( (radioSelection & 0xF) == 2 ? 2 : ( (radioSelection & 0xF) == 4 ? 3 : 4 ) );
	xil_printf("Applying TxDCO correction for radio %d\r\n", radioNum);

	//Mimic the radio test code, in hopes of a more stable EEPROM read...
	//Choose the EEPROM on the selected radio board; second arg is [0,1,2,3,4] for [FPGA, radio1, radio2, radio3, radio4]
	eepromStatus = WarpEEPROM_EEPROMSelect((unsigned int *)XPAR_EEPROM_0_MEM0_BASEADDR, 0);
	if(eepromStatus != 0)
	{
		xil_printf("EEPROM Select Failed!\r\n");
		return;
	}
	
	//Initialize the EEPROM controller
	eepromStatus = WarpEEPROM_Initialize((unsigned int *)XPAR_EEPROM_0_MEM0_BASEADDR);
	if(eepromStatus != 0)
	{
		xil_printf("EEPROM Init Returned %x\r\n", eepromStatus);
		xil_printf("EEPROM Init Failed!\r\n");
	return;
	}

	//Select the EEPROM on the current radio board
	eepromStatus = WarpEEPROM_EEPROMSelect((unsigned int*)XPAR_EEPROM_0_MEM0_BASEADDR, radioNum);

	if(eepromStatus != 0)
	{
		xil_printf("TxDCO: EEPROM error\r\n");
		return -1;
	}
	
	//Read the first page from the EERPOM
	WarpEEPROM_ReadMem((unsigned int*)XPAR_EEPROM_0_MEM0_BASEADDR, 0, 0, memory);
	version = (memory[0] & 0xE0) >> 5;
	revision = (memory[1] & 0xE0) >> 5;
	valid = memory[1] & 0x1F;

	xil_printf("\r\n\r\nEEPROM Values for Radio Board in Slot %d\r\n", radioNum);

	xil_printf("    WARP Radio Board Version %d.%d\r\n", version, revision);

	serial = WarpEEPROM_ReadWARPSerial((unsigned int*)XPAR_EEPROM_0_MEM0_BASEADDR);

	xil_printf("    Serial Number (WARP): WR-a-%05d\r\n", serial);

	WarpEEPROM_ReadDSSerial((unsigned int*)XPAR_EEPROM_0_MEM0_BASEADDR, memory);
	print("    EEPROM Hard-wired Serial Number: ");
	for(i=1;i<7;i++)
		xil_printf(" %x",memory[7-i]);
	xil_printf("\r\n\r\n");
	//Read the Tx DCO values
	calReadback = WarpEEPROM_ReadRadioCal((unsigned int*)XPAR_EEPROM_0_MEM0_BASEADDR, 2, 1);
	
	//Scale the stored values
	best_I = (signed short)(((signed char)(calReadback & 0xFF))<<1);
	best_Q = (signed short)(((signed char)((calReadback>>8) & 0xFF))<<1);
	
	xil_printf("TxDCO: Applied values to radio %d - I: %d\tQ: %d\r\n", radioNum, best_I, best_Q);
	
	//Finally, write the Tx DCO values to the DAC
	WarpRadio_v1_DACOffsetAdj(ICHAN, best_I, radioSelection);
	WarpRadio_v1_DACOffsetAdj(QCHAN, best_Q, radioSelection);
	
	return 0;
}


/******* AGC core control functions ********/
void ofdm_AGC_SetDCO(unsigned int AGCstate){
	
	// Enables DCO and DCO subtraction (correction scheme and butterworth hipass are active)
	
	unsigned int bits;
	
	bits = OFDM_AGC_MIMO_ReadReg_Bits(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR);
	
	if(AGCstate)
		bits = bits | 0x6;
	else
		bits = bits & 0x1;
	
	OFDM_AGC_MIMO_WriteReg_Bits(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, bits);
	
	return;
}

void ofdm_AGC_Reset(){
	
	// Cycle the agc's software reset port
	
	OFDM_AGC_MIMO_WriteReg_SRESET_IN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 1);
	usleep(10);
	OFDM_AGC_MIMO_WriteReg_SRESET_IN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0);
	usleep(100);
	
	return;
}


void ofdm_AGC_MasterReset(){
	
	// Cycle the master reset register in the AGC and enable it
	
	OFDM_AGC_MIMO_WriteReg_AGC_EN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0);
	usleep(10);
	OFDM_AGC_MIMO_WriteReg_MRESET_IN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0);
	usleep(10);
	OFDM_AGC_MIMO_WriteReg_MRESET_IN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 1);
	usleep(10);
	OFDM_AGC_MIMO_WriteReg_MRESET_IN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0);
	usleep(10);
	OFDM_AGC_MIMO_WriteReg_AGC_EN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 1);
	
	return;
}

void ofdm_AGC_Initialize(int noise_estimate){
	
	int g_bbset = 0;
	
	// First set all standard parameters
	
	// Turn off both resets and the master enable
	OFDM_AGC_MIMO_WriteReg_AGC_EN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0);
	OFDM_AGC_MIMO_WriteReg_SRESET_IN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0);
	OFDM_AGC_MIMO_WriteReg_MRESET_IN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0);
	
	// An adjustment parameter
	OFDM_AGC_MIMO_WriteReg_ADJ(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 8);
	
	// Timing for the DC-offset correction
	OFDM_AGC_MIMO_WriteReg_DCO_Timing(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0x46403003);
	
	// Initial baseband gain setting
	OFDM_AGC_MIMO_WriteReg_GBB_init(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 52);
	
	// RF gain AGCstate thresholds
	OFDM_AGC_MIMO_WriteReg_Thresholds(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 
		((AGC_THRESH_1&0xFF)<<16) +
		((AGC_THRESH_2&0xFF)<<8) + 
		 (AGC_THRESH_3&0xFF)
	);
	
	// Overall AGC timing
	OFDM_AGC_MIMO_WriteReg_Timing(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0x9A962A28);//0x826E3C0A;
	
	// vIQ and RSSI average lengths
	OFDM_AGC_MIMO_WriteReg_AVG_LEN(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0x10F); //103
	
	// Disable DCO, disable DCO subtraction, set filter to straight downsampling
	OFDM_AGC_MIMO_WriteReg_Bits(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, 0x0);
	
	// Compute and set the initial g_BB gain value from the noise estimate
	// The initial g_bb sets noise to -19 db, assuming 32 db RF gain
	
	g_bbset = -19 - 32 - noise_estimate;
	OFDM_AGC_MIMO_WriteReg_GBB_init(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, g_bbset);
	
	// Perform a master reset
	ofdm_AGC_MasterReset();
	
	// Agc is now reset and enabled, ready to go!
	return;
}


void ofdm_AGC_setNoiseEstimate(int noise_estimate){
	int g_bbset;
	
	g_bbset = -19 - 32 - noise_estimate;
	
	OFDM_AGC_MIMO_WriteReg_GBB_init(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, g_bbset);
	
	return;
}

unsigned int ofdm_AGC_GetGains(void){
	
	unsigned int gBB_A, gRF_A, gBB_B, gRF_B, gains;
	
	// Get the gains from the registers
	gBB_A = OFDM_AGC_MIMO_ReadReg_GBB_A(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR);
	gRF_A = OFDM_AGC_MIMO_ReadReg_GRF_A(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR);
	
	gBB_B = OFDM_AGC_MIMO_ReadReg_GBB_B(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR);
	gRF_B = OFDM_AGC_MIMO_ReadReg_GRF_B(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR);
	
	// First concatenate the two radios together, into the gRF register
	// 2 lowest bits are RF, 5 higher bits are BB, last bit is unused
	// Multiply by 2^2, shift gBB right by 2 bits
	
	gRF_A = gRF_A + (gBB_A * 4);
	gRF_B = gRF_B + (gBB_B * 4);
	
	// Multiply by 2^8 shift gRF right by 8 bits
	gains = gRF_A + (gRF_B * 256);
	
	// Returns hi[0 gBB_B g_RF_B | 0 g_BB_A g_RF_A]lo
	return gains;
}

void ofdm_AGC_SetTarget(unsigned int target){
	OFDM_AGC_MIMO_WriteReg_T_dB(XPAR_OFDM_AGC_MIMO_OPBW_0_BASEADDR, target);
	return;
}
/******* END AGC core control functions ********/



/******* WARP Timer core control functions *******/
void warp_timer_start(unsigned char timer) {
	warp_timer_WriteReg_control(warp_timer_ReadReg_control() | (TIMER_MASK_CALC(timer) & TIMER_CONTROL_START_MASK));
	warp_timer_WriteReg_control(warp_timer_ReadReg_control() & ~(TIMER_MASK_CALC(timer) & TIMER_CONTROL_START_MASK));
}

void warp_timer_pause(unsigned char timer) {
	warp_timer_WriteReg_control(warp_timer_ReadReg_control() | (TIMER_MASK_CALC(timer) & TIMER_CONTROL_PAUSE_MASK));
}

void warp_timer_resume(unsigned char timer) {
	warp_timer_WriteReg_control(warp_timer_ReadReg_control() & ~(TIMER_MASK_CALC(timer) & TIMER_CONTROL_PAUSE_MASK));
}

void warp_timer_setMode(unsigned char timer, unsigned char mode) {
	if(mode == 1)
		warp_timer_WriteReg_control(warp_timer_ReadReg_control() | (TIMER_MASK_CALC(timer) & TIMER_CONTROL_MODE_MASK));
	else
		warp_timer_WriteReg_control(warp_timer_ReadReg_control() & ~(TIMER_MASK_CALC(timer) & TIMER_CONTROL_MODE_MASK));
}

void warp_timer_resetDone(unsigned char timer) {
	warp_timer_WriteReg_control(warp_timer_ReadReg_control() | (TIMER_MASK_CALC(timer) & TIMER_CONTROL_RESETDONE_MASK));
	warp_timer_WriteReg_control(warp_timer_ReadReg_control() & ~(TIMER_MASK_CALC(timer) & TIMER_CONTROL_RESETDONE_MASK));
}

void warp_timer_resetAllDoneStatus() {
	warp_timer_WriteReg_control(warp_timer_ReadReg_control() | (TIMER_CONTROL_RESETDONE_MASK));
	warp_timer_WriteReg_control(warp_timer_ReadReg_control() & ~(TIMER_CONTROL_RESETDONE_MASK));
}

void warp_timer_setTimer(unsigned char timer, unsigned int slotTime, unsigned int slotCount) {
	switch(timer) {
	case 0:
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER0_SLOTCOUNT, (Xuint32)(slotCount));
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS01_SLOTTIME, ( (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS01_SLOTTIME) & 0x0000FFFF) | slotTime) );
		return;
	case 1:
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER1_SLOTCOUNT, (Xuint32)(slotCount));
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS01_SLOTTIME, ( (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS01_SLOTTIME) & 0xFFFF0000) | (slotTime<<16) ) );
		return;
	case 2:
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER2_SLOTCOUNT, (Xuint32)(slotCount));
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS23_SLOTTIME, ( (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS23_SLOTTIME) & 0x0000FFFF) | slotTime) );
		return;
	case 3:
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER3_SLOTCOUNT, (Xuint32)(slotCount));
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS23_SLOTTIME, ( (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS23_SLOTTIME) & 0xFFFF0000) | (slotTime<<16) ) );
		return;
	case 4:
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER4_SLOTCOUNT, (Xuint32)(slotCount));
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS45_SLOTTIME, ( (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS45_SLOTTIME) & 0x0000FFFF) | slotTime) );
		return;
	case 5:
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER5_SLOTCOUNT, (Xuint32)(slotCount));
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS45_SLOTTIME, ( (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS45_SLOTTIME) & 0xFFFF0000) | (slotTime<<16) ) );
		return;
	case 6:
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER6_SLOTCOUNT, (Xuint32)(slotCount));
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS67_SLOTTIME, ( (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS67_SLOTTIME) & 0x0000FFFF) | slotTime) );
		return;
	case 7:
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER7_SLOTCOUNT, (Xuint32)(slotCount));
		XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS67_SLOTTIME, ( (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMERS67_SLOTTIME) & 0xFFFF0000) | (slotTime<<16) ) );
		return;
	default:
		return;
	}
}

unsigned char warp_timer_getStatus(unsigned char timer) {
	return (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER_STATUS) >> (timer*4)) & 0xf;
}

unsigned char warp_timer_isDone(unsigned char timer) {
	return (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER_STATUS) & TIMER_MASK_CALC(timer) & TIMER_STATUS_DONE_MASK) != 0;
}

unsigned char warp_timer_isActive(unsigned char timer) {
	return (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER_STATUS) & TIMER_MASK_CALC(timer) & TIMER_STATUS_RUNNING_MASK) != 0;
}

unsigned char warp_timer_isPaused(unsigned char timer) {
	return (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER_STATUS) & TIMER_MASK_CALC(timer) & TIMER_STATUS_PASUED_MASK) != 0;
}

unsigned int warp_timer_getStatuses() {
	return XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER_STATUS);
}

unsigned char warp_timer_getDoneStatus() {
	return (XIo_In32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER_STATUS) & TIMER_STATUS_DONE_MASK);
}

void warp_timer_init(){
	//Clear all start/pause/mode/donereset bits
	XIo_Out32(XPAR_WARP_TIMER_PLBW_0_MEMMAP_TIMER_CONTROL, 0);

	//Set sane values for the slot times and slot counts
	warp_timer_setTimer(0, 500, 32);
	warp_timer_setTimer(1, 500, 32);
	warp_timer_setTimer(2, 500, 32);
	warp_timer_setTimer(3, 500, 32);
	warp_timer_setTimer(4, 500, 32);
	warp_timer_setTimer(5, 500, 32);
	warp_timer_setTimer(6, 500, 32);
	warp_timer_setTimer(7, 500, 32);

	//Clear any stale timer-done status bits
	warp_timer_resetAllDoneStatus();
	return;
}
/******* END WARP Timer core control functions *******/

///@brief Set the transmit power (via the radio's RF VGA)
///@param txPwr Desired transmit power; must be integer in [0,63]
///@return Returns -1 if an invalid power is requested; 0 on success
int warpphy_setTxPower(unsigned char txPwr)
{
	if(txPwr > 63)
		return -1;
	else
		WarpRadio_v1_SetTxGainTiming(FIRST_RADIO | SECOND_RADIO, (0x3F & txPwr), 0xF, 1);

	return 0;
}

#ifdef CONVCODED_PHY
void warpphy_setDecodingMode(unsigned char mode)
{
	if(mode == DECODINGMODE_SOFT)
	{
		RxSoftDecodingEn = 1;
	}
	else if(mode == DECODINGMODE_HARD)
	{
		RxSoftDecodingEn = 0;
	}

	ofdm_txrx_mimo_WriteReg_FEC_Coding(OFDM_BASEADDR,
		(SOFT_DECODING * RxSoftDecodingEn) |
		(QPSK_SCL * 2^4) |
		(QAM16_SCL * 2^8)
		);

	return;
}
#endif

/*********************************************************************/
/* A whole bunch of debugging functions - these will go away someday */
/*********************************************************************/

#if INCLUDE_WARPPHY_DEBUG_FUNCTIONS

void warpphy_setPktDlyPlus(){
	ofdmRx_pktDetDly++;
	xil_printf("+ofdmRx_pktDetDly = %d\r\n", ofdmRx_pktDetDly);
	mimo_ofdmRx_setPktDetDly(ofdmRx_pktDetDly);
}
void warpphy_setPktDlyMinus(){
	ofdmRx_pktDetDly--;
	xil_printf("-ofdmRx_pktDetDly = %d\r\n", ofdmRx_pktDetDly);
	mimo_ofdmRx_setPktDetDly(ofdmRx_pktDetDly);
}

/****************************/
/* Phase noise tracking */
/****************************/
void warpphy_set_PN_KPlus(unsigned int increment){
	PN_KVal=PN_KVal+increment;
	mimo_ofdmRx_setPNTrack_K(PN_KVal);
	xil_printf("K: %x\r\n", PN_KVal);
}

void warpphy_set_PN_KMinus(unsigned int decrement){
	PN_KVal=PN_KVal-decrement;
	mimo_ofdmRx_setPNTrack_K(PN_KVal);
	xil_printf("K: %x\r\n", PN_KVal);
}

/*****************/
/* CFO Constants */
/*****************/

void print_CFO_constants ()
{
	xil_printf("AK: %x\tBP: %x\tBI: %x\r\n", A_KPval, B_KPval, B_KIval);
}

void warpphy_set_B_KPPlus(unsigned int increment){
	B_KPval=B_KPval+increment;
	mimo_ofdmRx_setCFO_B_KP(B_KPval);
	print_CFO_constants();
}

void warpphy_set_B_KPMinus(unsigned int decrement){
	B_KPval=B_KPval-decrement;
	mimo_ofdmRx_setCFO_B_KP(B_KPval);
	print_CFO_constants();
}

void warpphy_set_B_KIPlus(unsigned int increment){
	B_KIval=B_KIval+increment;
	mimo_ofdmRx_setCFO_B_KI(B_KIval);
	print_CFO_constants();
}

void warpphy_set_B_KIMinus(unsigned int decrement){
	B_KIval=B_KIval-decrement;
	mimo_ofdmRx_setCFO_B_KI(B_KIval);
	print_CFO_constants();
}

void warpphy_set_CFODebugOutput(unsigned char outputSel){
	
	if(outputSel)
		ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR) | DEBUG_CFO_OUTSEL);
	else
		ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR) & ~DEBUG_CFO_OUTSEL);
	
	xil_printf("CFO Debug Output %d\r\n",outputSel);
}


/*********************/
/* Other PHY options */
/*********************/

void warpphy_set_FFTOffset_Plus(){
	RxFFT_Window_Offset++;
	mimo_ofdmRx_setFFTWindowOffset(RxFFT_Window_Offset);
	xil_printf("Rx FFT Offset: %d\r\n", RxFFT_Window_Offset);
}

void warpphy_set_FFTOffset_Minus(){
	RxFFT_Window_Offset--;
	mimo_ofdmRx_setFFTWindowOffset(RxFFT_Window_Offset);
	xil_printf("Rx FFT Offset: %d\r\n", RxFFT_Window_Offset);
}

void warpphy_setNoiseTargetPlus(){
	agc_noisePowerEstimate++;
	xil_printf("+Noise Estimate = %d\r\n", agc_noisePowerEstimate);
	ofdm_AGC_setNoiseEstimate(agc_noisePowerEstimate);
}

void warpphy_setNoiseTargetMinus(){
	agc_noisePowerEstimate--;
	xil_printf("-Noise Estimate = %d\r\n", agc_noisePowerEstimate);
	ofdm_AGC_setNoiseEstimate(agc_noisePowerEstimate);
}

void warpphy_setTargetPlus(){
	agc_targetRxPowerOut++;
	xil_printf("+agc_targetRxPowerOut = %d\r\n", agc_targetRxPowerOut);
	ofdm_AGC_SetTarget(agc_targetRxPowerOut);
}

void warpphy_setTargetMinus(){
	agc_targetRxPowerOut--;
	xil_printf("-agc_targetRxPowerOut = %d\r\n", agc_targetRxPowerOut);
	ofdm_AGC_SetTarget(agc_targetRxPowerOut);
}

void warpphy_setPktDetPlus(unsigned int offset){
	pktDet_energyThresh=pktDet_energyThresh+offset;
	//ofdm_pktDetector_mimo_WriteReg_pktDet_avgThresh(PKTDET_BASEADDR, pktDet_energyThresh);
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setThresh(pktDet_energyThresh);
	xil_printf("pktDet_energyThresh = %d\r\n",pktDet_energyThresh);
}

void warpphy_setPktDetMinus(unsigned int offset){
	pktDet_energyThresh=pktDet_energyThresh-offset;
	//ofdm_pktDetector_mimo_WriteReg_pktDet_avgThresh(PKTDET_BASEADDR, pktDet_energyThresh);
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setThresh(pktDet_energyThresh);
	
	xil_printf("pktDet_energyThresh = %d\r\n",pktDet_energyThresh);
}

void warpphy_setPktDetMinDurPlus(unsigned int offset){
	pktDet_threshMinDur = pktDet_threshMinDur + offset;

	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setMinDuration(pktDet_threshMinDur);
	xil_printf("Minimum Duration = %d\r\n", pktDet_threshMinDur);
}

void warpphy_setPktDetMinDurMinus(unsigned int offset){
	pktDet_threshMinDur = pktDet_threshMinDur - offset;
	
	ofdm_txrx_mimo_WriteReg_Rx_PktDet_setMinDuration(pktDet_threshMinDur);
	xil_printf("Minimum Duration = %d\r\n", pktDet_threshMinDur);
}

void warpphy_setCSMAPlus(unsigned int offset){
	pktDet_carrierSenseThresh = pktDet_carrierSenseThresh + offset;

	ofdm_txrx_mimo_WriteReg_Rx_CSMA_setThresh(pktDet_carrierSenseThresh);
	xil_printf("Carrier Sense Thresh = %d\r\n", pktDet_carrierSenseThresh);
}

void warpphy_setCSMAMinus(unsigned int offset){
	pktDet_carrierSenseThresh = pktDet_carrierSenseThresh - offset;
    
	ofdm_txrx_mimo_WriteReg_Rx_CSMA_setThresh(pktDet_carrierSenseThresh);
	xil_printf("Carrier Sense Thresh = %d\r\n", pktDet_carrierSenseThresh);
}

void warpphy_setGainPlus(unsigned int offset){
	txGain=txGain+offset;
	WarpRadio_v1_SetTxGainTiming(FIRST_RADIO | SECOND_RADIO, txGain, 0xF, 2);
	xil_printf("TxGain = %x\r\n",txGain);
}

void warpphy_setGainMinus(unsigned int offset){
	txGain=txGain-offset;
	WarpRadio_v1_SetTxGainTiming(FIRST_RADIO | SECOND_RADIO, txGain, 0xF, 2);
	xil_printf("TxGain = %x\r\n",txGain);
}
#endif

